'''
Created on Jul 8, 2009

@author: yongjinp

time-varying hierarchical random graph model
'''

from numpy import sum, cumsum, exp, inf, prod, log, log10, isnan, inf, ones
from numpy import max as np_max

from scipy.special import digamma
import random
import time
import networkx as nx
from ds.fixed_tree import *
from inference.linkpred import *

# from graphio import *
# to measure speed
# import resource
# def cpu_time(): return resource.getrusage(resource.RUSAGE_SELF)[0]


class TimeData: # a collection of data at a certain time
    
    # data
    masking = None
    
    # latent variable (and related ones) 
    Z = None        
        
    def __init__(self, masking, KK):        
        self.masking = masking        
        NN = len( masking ); KK = KK;
        self.Z = zeros((NN,KK))  
        pass
        
    pass # End of class timeData




class TimeVaryingEngine:
    
    __slots__ = ['timeGraph','fullGraph','Edges','Adj','tdList','FBT',
                 'link','maxZ','avgZ','comember','NN','KK','lmd','vertices']
    
    timeGraph = None # a graph explaining relationship between time frames    
    dataGraphs = None # a full graph topology  
    Edges = None     # a list of overall edges
    Adj = None       # a dict of overall adj relations
    FBT = None       # a fixed binary tree
    vertices = None
    times = None
    
    tdList = None    # a list of time data
    link = None      # a list of link prediction matrix
    comember = None  # a list of co-membership matrix
    avgZ = None
    maxZ = None
    
    NN = 0.; KK = 0.;  # dim of a latent variable matrix
    T = 0.;            # total length of time
    
    def __init__(self,dGs,tG,depth,masking_mat=None,vertices=None):
        '''
        takes : dGs                  a list of data graphs
                tG                   time-dependency graph
                depth                depth of hierarchical model
                masking_mat=None     masking matrix
                vertices=None        names of vertices (string type)
                
        '''
        # data
        self.dataGraphs = dGs
        self.timeGraph = tG
                
        # names of times
        self.times = tG.nodes()
        self.times.sort()
                
        # amass graph data for easier look up
        if type(dGs)==type([]):     
            self.set_graph_data(dGs,vertices)
        elif type(dGs)==type({}):
            tmpGs = [dGs[tt] for tt in self.times]
            self.set_graph_data(tmpGs,vertices)
        
        
        NN, TT = len(self.vertices), tG.number_of_nodes()
        self.T = TT
        
        # build masking matrix based on the presence of vertex    
        if masking_mat==None:
            masking_mat = zeros((NN,TT))                    
            for ti, tt in enumerate( self.times ):
                snapshot = dGs[tt]
                for vi,vv in enumerate( self.vertices ):
                    if snapshot.has_node(vv) :
                        masking_mat[vi,ti] = 1.;
        
        
        # masking matrix should cover all vertices
        assert( masking_mat.shape[0]==NN )
        
        # num of possible clusters
        KK = 2**depth; self.KK = KK;     
        
        # time varying data
        self.tdList = {}
        for ti, tt in enumerate( self.times ):            
            self.tdList[tt] = TimeData(masking_mat[:][:,ti],KK)
        
        # data structure
        self.FBT = FixedBTree()
        self.FBT.init_bottom_up(depth)        
        self.FBT.init_params(self.T)
        pass
    
    #---- #--- Learners ---#
    ## with Stochastic assignment
    def learn_stoc(self,maxiter,burnin,lmd=None,interval=1,eta=1e-3,warmstart=False,verbose=True):
        return self.learn('stoc', maxiter, lmd=lmd, eta=eta, burnin=burnin, 
                          interval=interval, warmstart=warmstart, verbose=verbose)
    
    ## with Mean approximation
    def learn_mean(self,maxiter,lmd=None,eta=1e-3,tol=1e-3,warmstart=False,verbose=True):
        return self.learn('mean', maxiter, lmd=lmd, eta=eta, tol=tol, warmstart=warmstart, verbose=verbose)
    
    ## with Max approximation
    def learn_max(self,maxiter,lmd=None,eta=1e-3,tol=1e-3,warmstart=False,verbose=True):
        return self.learn('max', maxiter, lmd=lmd, eta=eta, tol=tol, warmstart=warmstart, verbose=verbose)
    
    
    ## meanfield-approximated learning algorithms... 
    def learn(self,method,maxiter,lmd=0.,eta=1e-3,tol=1e-3,burnin=0,interval=1,warmstart=False,verbose=True):
        
        NN = self.NN; KK = self.KK; T = self.T; tdlist = self.tdList;
        converged = check_convergence
        times = self.times
        
        if method == 'mean': latent_update = latent_update_mean
        elif method == 'max':latent_update = latent_update_max
        else:                latent_update = latent_update_stoc
                
        # result accumulators
        maxZ = []; avgZ = [zeros((NN,KK)) for tt in xrange(T)]
        # initialize posterior link prediction
        self.init_link()
        
        # temporary data holders    
        lnprob0 = zeros((1,KK)); lnprob = zeros((1,KK))
                
        if not(warmstart):
            if verbose: print 'Random Initialization...'
            for tt in xrange(T):
                set_latent_random( tdlist[times[tt]].Z )
        else:
            if verbose: print 'Warm Initialization...'
        
        maxllik = sum(self.update_tree_mf(tt) for tt in xrange(T))
        llikvec = [maxllik]
        for iter in xrange(maxiter):
            
            if verbose: print iter,
            llik = 0.; # cgrad = 0.
            for tt in xrange(T):
                td = tdlist[times[tt]]
                
                if verbose: print '.',             
                for vidx in xrange(NN):                  
                    if not(td.masking[vidx]==1): # this vertex is not active
                        latent_update_mean(td.Z,vidx,lnprob0)
                        continue
                    
                    for kk in xrange(KK):
                        lnprob[0,kk] = self.path_likelihood(vidx, tt, kk, lmd)
                    
                    latent_update(td.Z,vidx,lnprob)
                    pass # End of vertex update                                                
                                
                if method == 'mean':                
                    llik += self.update_tree_mf(tt);
                else:
                    llik += self.update_tree_cd(tt,eta);
                
                pass # End of the iteration of each time slots
            
            # Accumulate summary statistics?             
            if method=='stoc' and iter>burnin and iter%interval==0: 
                for tt in xrange(T): avgZ[tt] += tdlist[times[tt]].Z # average membership
                linkupdate(self) # co-membership and link prediction
            
            if llik > maxllik :
                if verbose: print 'm',
                maxZ = []
                for tt in xrange(T):
                    maxZ.append( tdlist[times[tt]].Z.copy() )                
                maxllik = llik
                        
            if verbose: print 'llik=%.2f'%(llik)
            llikvec.append(llik) # total log-likelihood vector
            
            if not(method=='stoc') and converged(llikvec,iter,tol): break
            
            pass # End of update / sampling iterations
        
        
        # normalize summary statistics
        if method == 'stoc':
            for tt in xrange(T): avgZ[tt] /= (maxiter-burnin)
            self.avgZ = avgZ
            linknormalize( self,(maxiter-burnin) )            
        else:
            linkupdate( self, cont=True )
            self.avgZ = []
            for tt in xrange(T): self.avgZ.append( tdlist[times[tt]].Z.copy() )                
            pass
        
        self.maxZ = maxZ
        
        return llikvec
    
        
    #--- #--- Inference method ---#        
    def path_likelihood(self,vidx,tt,kk,lmd=0.,eps=1e-20):
        '''
        Inference of latent variables P( Z[vidx,kk]==1 ) at time tt
        '''
        
        fbt = self.FBT
        leaves = fbt.leaves()
        
        Adj = self.Adj
        Edges = self.Edges
                
        # time graph data
        tdlist = self.tdList
        times = self.times
        tdata = tdlist[times[tt]]
        mm = tdata.masking
        Z = tdata.Z
        
        assert( mm[vidx]==1 )
                        
        ### 1. collapsed model
        nn = leaves[kk]
        
        # valid partners connecting with vidx        
        validedge = Adj[tt][vidx]
        Eopp = sum(Z[validedge,kk])
        Etotopp = nn.Nr[tt] - Z[vidx,kk] # the rest of members in the group
        
        ltheta1, ltheta0 = nn.lnTheta1[tt], nn.lnTheta0[tt]
        rho = ltheta1-ltheta0
        
        # accumulate log-likelihood
        llik = rho*Eopp + ltheta0*Etotopp
                        
        ### 2. left-right model
        while nn != fbt.root:
            Pa = fbt.parent(nn)
            opposite = fbt.terminals_under_sibling(nn)
            # num of edges connecting opposite side            
            Eopp = sum(Z[validedge,:][:,opposite])
            
            # num of total edges connecting with the opposite side
            Etotopp = sum( leaves[kopp].Nr[tt] for kopp in opposite )
            
            ltheta1,ltheta0 = Pa.lnTheta1[tt],Pa.lnTheta0[tt] 
            rho = ltheta1-ltheta0
            
            # log-likelihood
            llik += rho*Eopp + ltheta0*Etotopp
            nn = Pa
            pass
        
        # there is time-constraint
        if lmd > 0.:
            tgraph = self.timeGraph            
            tadj_list = tgraph.neighbors(times[tt]);
            num = 0.; denom = 0.;
                        
            for tadj in tadj_list:
                td = tdlist[tadj];
                if not(td.masking[vidx]==1) : continue                
                
                muadj = max(td.Z[vidx,kk],eps) # Avoid a numerical error                
                num += log(muadj)              # thus, not so exact...
                denom += 1.
                pass
            # harmonic average
            llik = (llik + lmd*num) / (1. + lmd*denom)
            pass
        
        return llik
    
    
    
    #--- #--- parameter update methods ---#
    '''
    TREE update methods, which perform:
        # update sufficient statistics
        # update parameters
        # and compute total likelihood    
    '''
    def update_tree_mf(self,tt):
        '''
        meanfield update of a tree based on 'Z' at time tt    
        '''
        times = self.times
        tdata = self.tdList[times[tt]]
        Z = tdata.Z
        mm = tdata.masking
        
        fbt = self.FBT
        Edges = self.Edges
        NN = self.NN
        
        llik = 0.; # total log-likelihood
        for nn in fbt.dfs_traversal():
            
            # leaf node, collapsed model            
            if len( fbt.children(nn) ) == 0:
                kk = nn.idx            
                Nr = sum(Z[ii,kk] for ii in xrange(NN) if mm[ii]==1)                
                if Nr >= 2.:
                    Etot = Nr*(Nr-1.)/2.;
                    E = sum( Z[v1,kk]*Z[v2,kk] for v1,v2 in Edges[tt] ) /2.
                    
                else:
                    # not much advantage to compute this exactly as follows...
                    # Etot = sum( Z[v1,kk]*Z[v2,kk] for v1 in xrange(NN) for v2 in xrange(NN) 
                    #            if not(v1==v2) and mm[v1]==1 and mm[v2]==1 ) / 2.
                    # E = sum( Z[v1,kk]*Z[v2,kk] for v1,v2 in Edges[tt] ) /2.
                    
                    # so just set this case '0', prior will take care of smoothing
                    Etot = 0.; E = 0.;
                    
            else: # has children, left-right model
                
                Nr = sum( [cc.Nr[tt] for cc in fbt.children(nn)] )                
                left = fbt.left_terminals_under(nn)
                right = fbt.right_terminals_under(nn)
                
                Etot = prod( [cc.Nr[tt] for cc in fbt.children(nn)] )
                E = sum( Z[v1,ll]*Z[v2,rr] for v1,v2 in Edges[tt]
                         for ll in left for rr in right )
                
                pass
                        
            nn.Etot[tt] = Etot
            nn.E[tt] = E
            nn.Nr[tt] = Nr
            # parameter update and likelihood computation            
            ltheta1 = digamma(E+1.) - digamma(Etot+2.)
            ltheta0 = digamma(Etot-E+1.) - digamma(Etot+2.)
            
            nn.lnTheta1[tt], nn.lnTheta0[tt] = ltheta1, ltheta0
            nn.theta[tt] = float(E+1.) / float(Etot+2.)
            
            # this node's log-likelihood
            lliktmp = ltheta1*(E+1.) + ltheta0*(Etot-E+1.)
            
            llik += lliktmp
            pass ## End of DFS traversal
        
        return llik
            
    
    # gradient-based update... might be useful to get away from local optima    
    def update_tree_cd(self,tt,eta = 5e-4):
        '''
        optimize by the gradient of contrastive divergence at time tt                
        '''
        times = self.times
        tdata = self.tdList[times[tt]]
        
        Z = tdata.Z; mm = tdata.masking        
        fbt = self.FBT; Edges = self.Edges; NN = self.NN        
        
        llik = 0.; # total log-likelihood
        # cgrad = 0.; # cumulative gradient
        for nn in fbt.dfs_traversal():
            # previous edge counts
            Etot_old, E_old, theta_old = nn.Etot[tt], nn.E[tt], nn.theta[tt]
            # leaf node, collapsed model            
            if len( fbt.children(nn) ) == 0:
                kk = nn.idx; Nr = sum(Z[ii,kk] for ii in xrange(NN) if mm[ii]==1)
                
                # new sufficient statistics
                if Nr >= 1.:
                    Etot = Nr*(Nr-1.)/2.;
                    E = sum( Z[v1,kk]*Z[v2,kk] for v1,v2 in Edges[tt] ) /2.
                    # if mm[v1]==1 and mm[v2]==1 
                else:
                    # not much advantage to compute this exactly...
                    Etot = sum( Z[v1,kk]*Z[v2,kk] for v1 in xrange(NN) for v2 in xrange(NN)
                                if not(v1==v2) and mm[v1]==1 and mm[v2]==1 ) / 2.;
                    E = sum( Z[v1,kk]*Z[v2,kk] for v1,v2 in Edges[tt] ) /2.
                    
                                                
            else: # has children, left-right model                                
                Nr = sum( [cc.Nr[tt] for cc in fbt.children(nn)] )
                left = fbt.left_terminals_under(nn)
                right = fbt.right_terminals_under(nn)
                
                Etot = prod( [cc.Nr[tt] for cc in fbt.children(nn)] )
                E = sum( Z[v1,ll]*Z[v2,rr] for v1,v2 in Edges[tt]
                         for ll in left for rr in right )                
                pass
            
            nn.Etot[tt] = Etot; nn.E[tt] = E; nn.Nr[tt] = Nr
                                    
            dE = (E-E_old); dEbar = (Etot-E) -(Etot_old-E_old);
            grad = dE/theta_old - dEbar/(1.-theta_old)            
            grad /= 10.**log10(abs(grad)+1e-10) # scale down
            
            theta = theta_old + eta*grad
            theta = min(1.-1e-10,theta); theta = max(1e-10,theta);            
                                    
            # parameter update and likelihood computation
            ltheta1 = log(theta); ltheta0 = log(1.-theta);
            nn.lnTheta1[tt] = ltheta1; nn.lnTheta0[tt] = ltheta0;
            nn.theta[tt] = theta            
            
            # this node's log-likelihood
            lliktmp = 0.
            if Nr >= 1.: lliktmp = ltheta1*(E+1.) + ltheta0*(Etot-E+1.)
            
            llik += lliktmp
            pass ## End of DFS traversal
        
        return llik
    
    ##
    ##
    ##
    
    #--- #--- house keeping stuff ---#    
    def init_link(self):
        self.link = []; NN = self.NN; ngraphs = self.T
        for gidx in xrange(ngraphs):
            self.link.append( zeros((NN,NN)) )
        return 0
    
    def export_latent_matrix(self,method='avg'):
        ret = {}
        if method=='avg': matlist = self.avgZ
        elif method=='max': matlist = self.maxZ
        
        for tt,zz in enumerate(matlist):
            ret['latent%d'%tt] = zz.copy()                
        return ret
    
    def import_latent_matrix(self, indict):
        avgZ = self.avgZ
        times = self.times
        tdlist = self.tdList
        
        if self.avgZ == None :
            self.avgZ = {}; avgZ = self.avgZ;
        
        for tt in xrange(self.T):
            tdlist[times[tt]].Z = indict['latent%d'%tt].copy()
            avgZ[tt] = indict['latent%d'%tt].copy()
        return 0
    
    def export_link(self):
        ret = {}
        for tt,ll in enumerate(self.link):
            ret['link%d'%tt] = ll.copy()
        return ret
    
    '''
    convert NetworkX graph for our purpose
    '''
    def set_graph_data(self,graphs,vertices=None):
        
        # 1. merge into overall vertices
        if vertices == None:            
            vertices = []
            for G in graphs:
                vertices += filter(lambda x: x not in vertices, G.nodes_iter())
        
        self.graphs = graphs
        # 2. convert each
        # I don't like string-based indexing...
        # convert it to the number-based one    
        Adj = []; Edges = []
        for G in graphs: 
            adj_one, edges_one = convert_graph(G,vertices)
            Adj.append(adj_one); Edges.append(edges_one)
        
        self.Adj = Adj
        self.Edges = Edges
        self.vertices = vertices
        self.NN = len(vertices)
        
        return 0
        
    
    pass # End of class TimeVaryingEngine


##########################################################
# helper functions
##########################################################
def check_convergence(llikvec,iter,tol=1e-3):
    
    if iter < 3: return False
    
    #num1 = llikvec[iter]+llikvec[iter-1]
    #num2 = llikvec[iter-2]+llikvec[iter-3]
    
    if abs(llikvec[iter]-llikvec[iter-1]) < tol:
        print 'Converged'
        return True
    if abs(llikvec[iter]-llikvec[iter-3]) < tol :
        print 'Converged (fluctuating in 3 steps)'
        return True
    if abs(llikvec[iter]-llikvec[iter-2]) < tol :
        print 'Converged (fluctuating in 2 steps)'
        return True
    
    return False


def set_latent_random(Z,seed=None):
    if seed == None:
        random.seed(int(1e5*time.clock()))
    else:
        random.seed(seed)
            
    NN, KK = Z.shape        
    for ii in xrange(NN):
        Z[ii,:] = 0.
        Z[ii,random.randint(0,KK-1)] = 1.
    return Z




'''
Latent variable update functions
'''
def latent_update_stoc(Z,vidx,lnprob):
    prob = exp(lnprob - np_max(lnprob))
    cprob = cumsum(prob)
    cprob /= cprob[-1]
    
    uu = random.random()            
    newkk = sum(cprob<uu)
    #print oldkk, newkk
    #print cprob
    
    Z[vidx,:]=0.; Z[vidx,newkk]=1.;
            
    return 0

def latent_update_mean(Z,vidx,lnprob):
    prob0 = exp(lnprob - np_max(lnprob))    
    prob0 /= sum( prob0 )
    
    # prune out very small quantities
    K = float(Z.shape[1])    
    prob0[0,prob0[0,:]<1./K] = 0.
    prob = prob0 / sum(prob0)
    
    Z[vidx,:] = prob
    return 0

def latent_update_max(Z,vidx,lnprob):    
    newkk = argmax(lnprob)    
    Z[vidx,:]=0.;Z[vidx,newkk] = 1.;
    return 0


'''
discretize stochastic matrix
'''
def discretize(Z,copy=False):
    
    NN,KK = Z.shape
        
    if copy:
        ret = zeros((NN,KK))
    else:
        ret = Z
            
    for ii in xrange(NN):
        maxk = argmax(Z[ii,:])
        
        ret[ii,:]=0.
        ret[ii,maxk] = 1.
    
    return ret


'''
Nx Graph converter for a bit faster lookup
'''
# internal function
# helper function to convert NX graph into adj dict and edge list
def convert_graph(G,vertices):            
    Edges = []
    Adj = {}
    for ii,vi in enumerate(vertices):
        Adj[ii] = [];
        for jj,vj in enumerate(vertices):
            if (ii!=jj) and G.has_edge(vi,vj): 
                Adj[ii].append(jj)
                Edges.append( (ii,jj) )
    
    return Adj, Edges


